// Gateway Library Client
#ifdef _IA64_
    #pragma warning(disable:4127)
#endif 

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define DEFAULT_PORT        "36895"
#define DEFAULT_PROTO       SOCK_STREAM
#define DEFAULT_BUFFER_LEN  8192
#define BUFSIZE             8192

DWORD WINAPI ServerThread(LPVOID lpParam);
char * GlpToGlc(char * argv);
char * GlcToGlp(char * argv);
BOOL   CALLBACK    TraceFileWriting(LPSTR       lpData1);

char  chBuf[DEFAULT_BUFFER_LEN];
DWORD dwRead, dwWritten; 
BOOL fSuccess; 
HANDLE hIn, hOutKill, hInKill, hOut; 

void __cdecl main(int argc, char ** argv){
    struct addrinfo  hints, 
               *results = NULL,
               *addrptr = NULL;
    WSADATA    wsaData;
    SOCKET     *server_sockets = NULL;
    HANDLE     *server_threads = NULL;
    char       hoststr[NI_MAXHOST],
               servstr[NI_MAXSERV];
    char       *interface = NULL,
               *port = DEFAULT_PORT;
    int        socket_type = DEFAULT_PROTO,
               address_family = AF_UNSPEC,
               socket_count = 0,
               retval,
               i;

    hIn = (HANDLE)atoi(argv[0]);
    hOut = (HANDLE)atoi(argv[1]);
    port = argv[2];
    hInKill = (HANDLE)atoi(argv[3]);
    hOutKill = (HANDLE)atoi(argv[4]);
    CloseHandle(hInKill);
    CloseHandle(hOutKill);

    if ((retval = WSAStartup(0x202,&wsaData)) != 0){
      sprintf(chBuf, "WSAStartup failed with error %d",retval);
      GlcToGlp(chBuf);
      WSACleanup();
    }

    if ( _strnicmp(port, "0", 1) == 0){
      GlcToGlp("Check the port number");
    }

    memset(&hints, 0, sizeof(hints));
    hints.ai_family   = address_family;
    hints.ai_socktype = socket_type;
    hints.ai_protocol = ((socket_type == SOCK_DGRAM) ? IPPROTO_UDP : IPPROTO_TCP);
    hints.ai_flags    = ((interface == NULL) ? AI_PASSIVE : 0);

    retval = getaddrinfo(interface, port, &hints, &results);

    if (retval != 0){
      sprintf(chBuf, "getaddrinfo failed: %d", retval);
      GlcToGlp(chBuf);
      goto cleanup;
    }

    if (results == NULL){
      sprintf(chBuf, "Unable to resolve interface %s", interface);
      GlcToGlp(chBuf);
      goto cleanup;
    }

    addrptr = results;

    while (addrptr){
      socket_count++;
      addrptr = addrptr->ai_next;
    }

    server_sockets = (SOCKET *)HeapAlloc(
                                        GetProcessHeap(), 
                                        HEAP_ZERO_MEMORY, 
                                        sizeof(SOCKET) * socket_count
                                        );
    if (server_sockets == NULL){
      sprintf(chBuf, "HeapAlloc failed: %d", GetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    for (i=0; i < socket_count ;i++)
        server_sockets[i] = INVALID_SOCKET;

    socket_count = 0;
    addrptr = results;
    server_sockets[socket_count] = socket(addrptr->ai_family, 
                                          addrptr->ai_socktype, 
                                          addrptr->ai_protocol
                                          );
    if (server_sockets[socket_count] == INVALID_SOCKET){
      sprintf(chBuf, "socket failed: %d", WSAGetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    retval = bind(server_sockets[socket_count],
                  addrptr->ai_addr,
                  (int)addrptr->ai_addrlen
                  );
    if (retval == SOCKET_ERROR){
      sprintf(chBuf, "bind failed: %d", WSAGetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    if (addrptr->ai_socktype == SOCK_STREAM){
      retval = listen(server_sockets[socket_count],
                      7
                      );
      if (retval == SOCKET_ERROR){
        sprintf(chBuf, "listen failed: %d", WSAGetLastError());
        GlcToGlp(chBuf);
        goto cleanup;
      }
    }

    retval = getnameinfo(addrptr->ai_addr,
                         (socklen_t)addrptr->ai_addrlen,
                         hoststr,
                         NI_MAXHOST,
                         servstr,
                         NI_MAXSERV,
                         NI_NUMERICHOST | NI_NUMERICSERV
                         );
    if (retval != 0){
      sprintf(chBuf,"getnameinfo failed: %d", retval);
      GlcToGlp(chBuf);
      goto cleanup;
    }
    sprintf(chBuf, "socket 0x%x bound to address %s and port %s",
            server_sockets[socket_count], hoststr, servstr);
    GlcToGlp(chBuf);

    socket_count++;

    addrptr = addrptr->ai_next;

    server_threads = (HANDLE *)HeapAlloc(GetProcessHeap(),
                                         HEAP_ZERO_MEMORY,
                                         sizeof(HANDLE) * socket_count
                                         );
    if (server_threads == NULL){
      sprintf(chBuf, "HeapAlloc failed: %d", GetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    i = 0;
    server_threads[i] = CreateThread(NULL,
                                     0,
                                     ServerThread,
                                     (LPVOID)server_sockets[i],
                                     0,
                                     NULL
                                     );
    if (server_threads[i] == NULL){
      sprintf(chBuf, "CreateThread failed: %d", GetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    retval = WaitForMultipleObjects(socket_count,
                                   server_threads,
                                   TRUE,
                                   INFINITE
                                   );
    if ((retval == WAIT_FAILED) || (retval == WAIT_TIMEOUT)){
      sprintf(chBuf, "WaitForMultipleObjects failed: %d", GetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    cleanup:

    if (results != NULL){
      freeaddrinfo(results);
      results = NULL;
    }

    if (server_sockets != NULL){
      for (i=0; i < socket_count ;i++){
        if (server_sockets[i] != INVALID_SOCKET)
                closesocket(server_sockets[i]);
        server_sockets[i] = INVALID_SOCKET;
      }
      HeapFree(GetProcessHeap(), 0, server_sockets);
      server_sockets = NULL;
    }

    if (server_threads != NULL){
      for (i=0; i < socket_count ;i++){
        if (server_threads[i] != NULL)
                CloseHandle(server_threads[i]);
        server_threads[i] = NULL;
      }
      HeapFree(GetProcessHeap(), 0, server_threads);
      server_threads = NULL;
    }
}

char * GlpToGlc(char * argv){
    char * String = HeapAlloc(GetProcessHeap(), 
                              HEAP_ZERO_MEMORY, 
                              BUFSIZE
                              );
    fSuccess = ReadFile(hIn, String, BUFSIZE, &dwRead, NULL); 
    if (! fSuccess || dwRead == 0) return ("GlpToGls failed");
    return(String);
}
char * GlcToGlp(char * argv){
    fSuccess = WriteFile(hOut, argv, strlen(argv), &dwWritten, NULL); 
    if (! fSuccess) return ("GlcToGlp failed");
    return ("GlcToGlp succed");       
}

DWORD WINAPI ServerThread(LPVOID lpParam){
    SOCKET           s,
                     sc = INVALID_SOCKET;
    SOCKADDR_STORAGE from;
    char             Buffer[DEFAULT_BUFFER_LEN],
                     servstr[NI_MAXSERV],
                     hoststr[NI_MAXHOST];
    int              socket_type,
                     retval,
                     fromlen,
                     bytecount;
    DWORD            dw;
    char * String = HeapAlloc(GetProcessHeap(), 
                              HEAP_ZERO_MEMORY, 
                              BUFSIZE
                              );

    s = (SOCKET) lpParam;

    fromlen = sizeof(socket_type);
    retval = getsockopt(s, SOL_SOCKET, SO_TYPE, (char *)&socket_type, &fromlen);
    if (retval == INVALID_SOCKET){
      sprintf(chBuf, "getsockopt(SO_TYPE) failed: %d", WSAGetLastError());
      GlcToGlp(chBuf);
      goto cleanup;
    }

    for (;;){
      fromlen = sizeof(from);
      if (sc != INVALID_SOCKET){
        ZeroMemory(Buffer, DEFAULT_BUFFER_LEN);
        bytecount = recv(sc, Buffer, DEFAULT_BUFFER_LEN, 0);
        if (bytecount == SOCKET_ERROR){
          sprintf(chBuf, "recv failed: %d", WSAGetLastError());
          GlcToGlp(chBuf);
          goto cleanup;
        }
        else if (bytecount == 0){
          retval = shutdown(sc, SD_SEND);
          if (retval == SOCKET_ERROR){
            sprintf(chBuf, "shutdown failed: %d", WSAGetLastError());
            GlcToGlp(chBuf);
            goto cleanup;
          }
          closesocket(sc);
          sc = INVALID_SOCKET;
        }
        else{
          GlcToGlp(Buffer);
          String = GlpToGlc(" ");
          if (strlen(String)>0){
            strcat(String, "*");
            bytecount =  send(sc, String , strlen(String), 0);
            if (bytecount == SOCKET_ERROR){
              sprintf(chBuf, "send failed: %d", WSAGetLastError());
              GlcToGlp(chBuf);
              goto cleanup;
            }
            sprintf(String, "wrote %d bytes", bytecount);
            GlcToGlp(String);
          }else GlcToGlp("The string length is zero");          
        }
      }
      else{
        sc = accept(s, (SOCKADDR *)&from, &fromlen);
        if (sc == INVALID_SOCKET){
          sprintf(chBuf, "accept failed: %d", WSAGetLastError());
          GlcToGlp(chBuf);
          goto cleanup;
        }
        retval = getnameinfo((SOCKADDR *)&from,
                             fromlen,
                             hoststr,
                             NI_MAXHOST,
                             servstr,
                             NI_MAXSERV,
                             NI_NUMERICHOST | NI_NUMERICSERV
                             );
        if (retval != 0){
          sprintf(chBuf, "getnameinfo failed: %d", retval);
          GlcToGlp(chBuf);
          goto cleanup;
        }
        sprintf(chBuf, "Accepted connection from host %s and port %s",
                   hoststr, servstr);
        GlcToGlp(chBuf);
      }
    }

    cleanup:

    if (sc != INVALID_SOCKET){
      closesocket(sc);
      sc = INVALID_SOCKET;
    }

    return 0;
}
BOOL     CALLBACK    TraceFileWriting(LPSTR       lpData1){
    DWORD lpNumberOfBytesWritten;
    LPCTSTR lpFileName;
    HANDLE OpenFileHandle;
    int LenghtOfString;
	lpFileName ="d:\\TraceFileWriting.txt"; 			
	OpenFileHandle = CreateFile(
			lpFileName,
			GENERIC_WRITE | GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL);
	SetFilePointer(
			OpenFileHandle,
			0,
			NULL,
			FILE_BEGIN
			);	
    LenghtOfString = strlen(lpData1);
	WriteFile(OpenFileHandle,
			lpData1,
			LenghtOfString,
			&lpNumberOfBytesWritten,
			NULL);
	CloseHandle(OpenFileHandle);
	return TRUE;
}
